/*
 * libDebug
 *
 * Copyright (C) 2000 Patrick Alken
 * This library comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this library is distributed.
 *
 * $Id: trace-x86.c,v 1.11 2001/04/07 20:18:10 cosine Exp $
 */

#include <stdio.h>
#include <assert.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ptrace.h>
#include <signal.h>
#include <unistd.h>

/*
 * Top-level includes
 */
#include "args.h"
#include "break.h"
#include "config.h"
#include "debug.h"

#if defined(FreeBSD) || defined(OpenBSD)
# include "bsd-x86.h"
#elif defined(Linux)
# include "linux-x86.h"
#endif

#include "sub-x86.h"
#include "trace-x86.h"

/*
 * libString includes
 */
#include "alloc.h"

static int x86TraceDebugProcess(int function, int *data);

/*
 * Local: number of last signal received
 */
static int                  LastSignal = 0;

/*
x86StartDebugProcess()
  Start execution of the debugged process (DebugPath).

Inputs: none

Return: 1 if successful
        0 if not

Side effects: DebugPid is assigned to the pid of the traced
              process
*/

int
x86StartDebugProcess()

{
  pid_t pid;
  int waitval;

  assert(DebugPath != 0);

  if (RedirectIO)
  {
    if (pipe(Pipes) == (-1))
      return (0);
  }

  pid = fork();
  if (pid == (-1))
  {
    if (RedirectIO)
    {
      close(Pipes[0]);
      close(Pipes[1]);
    }

    return (0);
  }
  else if (pid == 0)
  {
    /*
     * Child: allow parent to trace us
     */
    if (ptrace(PT_TRACE_ME, 0, 0, 0) != 0)
      exit(0);

    assert(DebugArgs && DebugArgs[0]);

    if (RedirectIO)
    {
      /*
       * Redirect child's stdout and stderr to the parent's pipe
       */
      dup2(Pipes[1], 1);
      dup2(Pipes[1], 2);

      /*
       * Redirect child's stdin to come from parent's pipe
       */
      /* dup2(Pipes[1], 0); */

      /*
       * Close file descriptors
       */
      close(Pipes[0]);
      close(Pipes[1]);
    }

    execv(DebugPath, DebugArgs);

    exit(0);
  }
  else
  {
    /*
     * Parent process
     */

    DebugPid = pid;

    /*
     * wait for child to stop (execv)
     */
    wait(&waitval);

    /*
     * Set InstructionPointer to the program's entry point
     */
    InstructionPointer = x86GetCurrentInstruction();
  }

  return (1);
} /* x86StartDebugProcess() */

/*
x86TraceDebugProcess()
  Trace through our process using the given function

Inputs: function - ptrace function (PT_STEP or PT_CONTINUE)
        data     - modified to contain data depending on what
                   happens to debug process

Return: 0 if something goes wrong
        1 if everything is successful
        2 if program stops due to a signal (signal num goes in data)
        3 if program encounters a breakpoint (brk pt num goes in data)
        4 if program terminates normally (exit status goes in data)
        5 if program writes to stdout or stderr and we are
          redirecting output - the calling function can read the
          output using GetDebugOutput()

Special note about breakpoints:
  If this function is invoked from x86SingleStepOver(), it is
possible that we will run into a temporary breakpoint marked
with BK_STEPOVER. Since this breakpoint was set by x86SingleStepOver()
and not by the user, it does not have a legitimate breakpoint number,
and so the special value of 0 will be placed into 'data' indicating
to x86SingleStepOver() that it's temporary breakpoint was reached.
*/

static int
x86TraceDebugProcess(int function, int *data)

{
  int sig;
  int ptfunc;
  int waitval;

  assert(DebugPid != NOPID);

  ProcessRunning = 1;

  ptfunc = function;

  while (1)
  {
    if (ptfunc == PT_CONTINUE)
    {
      /*
       * Enable all breakpoints
       */
      EnableBreakpoints();
    }

    /*
     * Release child and let it continue
     */
  #if 0
    printf("PTRACING WITH SIGNAL [%d]\n", LastSignal);
  #endif
    if (ptrace(ptfunc, DebugPid, CONTADDR, 0) != 0)
      return (0);

    if (RedirectIO)
    {
      fd_set readfds;
      struct timeval TimeOut;
      int sret;
      size_t nread;

      /*
       * Since we are redirecting the child's output to our
       * pipe, we must read from Pipes[0] while the child is
       * executing, so make sure we use a non-blocking wait call.
       */

      TimeOut.tv_sec = 0;
      TimeOut.tv_usec = 0L;

      for (;;)
      {
        FD_ZERO(&readfds);
        FD_SET(Pipes[0], &readfds);

        sret = select(FD_SETSIZE, &readfds, 0, 0, &TimeOut);
        if (sret > 0)
        {
        #if 0
          fprintf(stderr, "sret = %d\n", sret);
        #endif
          if (FD_ISSET(Pipes[0], &readfds))
          {
            nread = read(Pipes[0], DebugOutput, sizeof(DebugOutput) - 1);
            if (nread == (-1))
              return (0);

            DebugOutput[nread] = '\0';
            return (5);
          }
        } /* if (sret > 0) */

        if (waitpid(DebugPid, &waitval, WNOHANG | WUNTRACED) != 0)
          break; /* something happened */
      } /* for (;;) */
    }
    else
    {
      /*
       * Wait for child to stop
       */
      wait(&waitval);
    }

    /*
     * Clear the last signal
     */
    LastSignal = 0;

    InstructionPointer = x86GetCurrentInstruction();

    if (ptfunc == PT_CONTINUE)
    {
      /*
       * Restore original instructions in place of breakpoint
       * instructions
       */
      DisableBreakpoints();
    }

    if (WIFEXITED(waitval))
    {
      /*
       * Process exited normally
       */
      EndDebugProcess();

      *data = WEXITSTATUS(waitval);

      /*
       * When a program exits, it sends two signals to the parent -
       * one for the exit status and the other is a SIGCHLD, so
       * make sure we catch the second.
       */
      wait(&waitval);

      return (4);
    }
    else if (WIFSTOPPED(waitval))
    {
      sig = WSTOPSIG(waitval);

      if (sig == SIGTRAP)
      {
        unsigned long addr;
        struct Breakpoint *bptr;

        addr = x86GetCurrentInstruction();
        assert(addr != 0);

        if (ptfunc == PT_STEP)
        {
          if ((bptr = FindBreakpoint(addr)))
          {
            *data = bptr->number;
            CheckBreakpoint(bptr);

            return (3);
          }

          /*
           * If we get here, one of two things must be true.
           * 1) We received a SIGTRAP while in single-stepping
           *    mode, but we never set a breakpoint at this
           *    address.
           * 2) We set a breakpoint at the previous instruction,
           *    and used the continue function to hit it, which
           *    then single-stepped past it and now wants to
           *    continue normally.
           *
           * In either case, do nothing and let it drop to
           * the bottom of the loop to decide whether to drop
           * out of the loop (case 1), or set ptfunc to PT_CONTINUE
           * and keep going (case 2).
           */
        } /* if (ptfunc == PT_STEP) */
        else if (ptfunc == PT_CONTINUE)
        {
          /*
           * We stopped exactly one byte after the breakpoint
           * instruction - go back one
           */
          --addr;

          bptr = FindBreakpoint(addr);
          if (!bptr)
          {
            /*
             * It appears to be a SIGTRAP which we did not set -
             * this is rare, but possible.
             */
            *data = sig;
            LastSignal = sig;
            return (2);
          }

          /*
           * Set eip to the real address
           */
          x86SetCurrentInstruction(addr);

          if (bptr->ignorecnt > 0)
          {
            /*
             * Ignore this breakpoint for now and continue
             * running the process
             */
            --bptr->ignorecnt;

            /*
             * If we use PT_CONTINUE in the next ptrace call,
             * it will simply stop at the very same breakpoint
             * as it just did - so step past this breakpoint and
             * then use PT_CONTINUE.
             */
            ptfunc = PT_STEP;
            continue;
          }

          /*
           * We hit a legitimate breakpoint - store the breakpoint
           * number and return to the user. We want to increase
           * the ignore count here, so if the user wants to continue
           * the process, we will know that we previously hit
           * this breakpoint, so the next time this function is called,
           * it will know to step past this breakpoint and then
           * continue normally.
           */
          ++bptr->ignorecnt;

          if (bptr->flags & BK_STEPOVER)
            *data = 0;
          else
            *data = bptr->number;

          /*
           * Check to see if this breakpoint should be deleted
           */
          CheckBreakpoint(bptr);

          return (3);
        } /* if (ptfunc == PT_CONTINUE) */
      } /* if (sig == SIGTRAP) */
      else
      {
        *data = sig;
        LastSignal = sig;
        return (2);
      }
    }

    if ((ptfunc == PT_STEP) && (function == PT_CONTINUE))
    {
      /*
       * The purpose of the step was simply to step past the
       * last breakpoint - we can continue normally now, since
       * our original function was PT_CONTINUE.
       */
      ptfunc = PT_CONTINUE;
    }
    else
    {
      /*
       * Our original function was a PT_STEP - it is completed
       * now so break out of the loop.
       */
      break;
    }
  } /* while (1) */

  /*
   * everything went well
   */
  return (1);
} /* x86TraceDebugProcess() */

/*
x86SingleStepInto()
  Single step our program, stepping into any subroutines

Inputs: num  - number of instructions to step through
        data - modified to contain info depending on the return
               result, see below

Return: 0 if unsuccessful
        1 if successful
        2 if program stops due to a signal
        3 if breakpoint is encountered
        4 if program terminates

Side effects: If the process stops due to a signal, data is
              modified to contain the signal number.
              If the process hits a breakpoint, the breakpoint
              number is put into data
              If the process terminates normally, data is modified
              to contain the exit status.
*/

int
x86SingleStepInto(int num, int *data)

{
  int ii;
  int ret;

  assert(num > 0);

  if (DebugPid == NOPID)
  {
    if (x86StartDebugProcess() == 0)
      return (0); /* something went wrong */
  }

  for (ii = 0; ii < num; ++ii)
  {
    ret = x86TraceDebugProcess(PT_STEP, data);
    if (ret != 1)
    {
      /*
       * Something happened (signal, breakpoint, exit, etc)
       */
      return (ret);
    }
  } /* for (ii = 0; ii < num; ++ii) */

  /*
   * All went well
   */
  return (1);
} /* x86SingleStepInto() */

/*
x86SingleStepOver()
  Single step our program, stepping past any subroutines

Inputs: num  - number of instructions to step over
        data - modified to contain info depending on the return
               result, see below

Return: 0 if unsuccessful
        1 if successful
        2 if program stops due to a signal
        3 if breakpoint encountered
        4 if program terminates

Side effects: If the process stops due to a signal, data is
              modified to contain the signal number.
              If the process stops due to a breakpoint, data is
              modified to contain the breakpoint number.
              If the process terminates normally, data is modified
              to contain the exit status.
*/

int
x86SingleStepOver(int num, int *data)

{
  int ii,
      dret,           /* return value from dump memory routine */
      tret;           /* return value from x86TraceDebugProcess() */
  unsigned long slen; /* length of subroutine call opcode */
  char *opbuf;        /* contains potential call opcode */

  assert(num > 0);

  if (DebugPid == NOPID)
  {
    if (x86StartDebugProcess() == 0)
      return (0); /* something went wrong */
  }

  for (ii = 0; ii < num; ++ii)
  {
    slen = 0;
    opbuf = 0;

    /*
     * Dump 20 bytes of memory so we can tell if we are about
     * to enter a subroutine
     */
    dret = x86DumpMemory(&opbuf, InstructionPointer, 20);

    if (opbuf)
    {
      if (dret > 0)
        slen = IsSubroutine(opbuf);

      MyFree(opbuf);
    }

    if (slen)
    {
      int bret;
      unsigned long baddr;

      /*
       * We are about to enter a subroutine - set a breakpoint
       * at location InstructionPointer + slen.
       */
      baddr = (unsigned long) (InstructionPointer + slen);
      bret = SetBreakpoint(baddr, BK_TEMPORARY | BK_STEPOVER);

      /*
       * Now continue until we hit the breakpoint
       */
      tret = x86TraceDebugProcess(PT_CONTINUE, data);

      if (tret == 3)
      {
        if (*data == 0)
        {
          /*
           * A value of 0 means we hit the temporary breakpoint
           * at the address after the call instruction - so we
           * do not need to inform the user about it. Continue
           * stepping.
           */
          continue;
        }

        /*
         * It is most likely a breakpoint which was set inside
         * of the subroutine we were trying to step over. It
         * could also be a breakpoint at the exact address we set
         * our temporary breakpoint.
         */
        return (3);
      }
      else
      {
        /*
         * 'data' will have been set appropriately by
         * x86ContinueDebugProcess()
         */
        return (tret);
      }
    } /* if (slen) */

    /*
     * We are not about to enter a subroutine - just step
     * one instruction
     */
    tret = x86TraceDebugProcess(PT_STEP, data);
    if (tret != 1)
    {
      /*
       * Something happened (signal, breakpoint, exit, etc)
       */
      return (tret);
    }
  } /* for (ii = 0; ii < num; ++ii) */

  /*
   * All went well
   */
  return (1);
} /* x86SingleStepOver() */

/*
x86ContinueDebugProcess()
  Continue the debugged process where it left off

Inputs: data - modified to contain info depending on return value

Return: 0 if unsuccessful
        1 if successful
        2 if program stops due to a signal (signal put into data)
        3 if breakpoint encountered (brk pt number put into data)
        4 if program terminates (exit status put into data)
*/

int
x86ContinueDebugProcess(int *data)

{
  if (DebugPid == NOPID)
  {
    if (x86StartDebugProcess() == 0)
      return (0); /* something went wrong */
  }

  return (x86TraceDebugProcess(PT_CONTINUE, data));
} /* x86ContinueDebugProcess() */

/*
x86KillDebugProcess()
  Kill current debugged process

Return: 1 if successful
        0 if not
*/

int
x86KillDebugProcess()

{
  int ret;
  int waitval;

  if (DebugPid != NOPID)
  {
    ret = ptrace(PT_KILL, DebugPid, 0, 0);

    if (ret == 0)
      wait(&waitval);
  }

  return (1);
} /* x86KillDebugProcess() */

/*
x86SetBreakpoint()
  Set a break point at memory location

Inputs: bptr - breakpoint structure containing info

Return: 1 if successful
        0 if not
*/

int
x86SetBreakpoint(struct Breakpoint *bptr)

{
  int saved; /* saved instruction */

  assert(bptr != 0);

  if (DebugPid == NOPID)
  {
    if (x86StartDebugProcess() == 0)
      return (0); /* something went wrong */
  }

  /*
   * Save lowest byte of the dword at the break address
   */
  saved = PtraceRead(DebugPid, bptr->address, 0);
  if (saved == (-1))
    return (0); /* error - most likely EIO */

  bptr->svdinsn = saved;

  return (1);
} /* x86SetBreakpoint() */

/*
x86EnableBreakpoint()
  Enable breakpoint by setting the first byte of it's
memory address to the breakpoint instruction

Return: 1 if successful
        0 if not
*/

int
x86EnableBreakpoint(struct Breakpoint *bptr)

{
  assert(DebugPid != NOPID);
  assert(bptr != 0);

  /*
   * Replace the instruction with our breakpoint instruction
   */
  if (PtraceWrite(DebugPid, bptr->address, BRKPT_INSN) != 0)
    return (0);

  return (1);
} /* x86EnableBreakpoint() */

/*
x86DisableBreakpoint()
  Restore breakpoint address with it's original instruction

Return: 1 if successful
        0 if not
*/

int
x86DisableBreakpoint(struct Breakpoint *bptr)

{
  assert(DebugPid != NOPID);
  assert(bptr != 0);

  /*
   * Replace the instruction with our saved instruction
   */
  if (PtraceWrite(DebugPid, bptr->address, bptr->svdinsn) != 0)
    return (0);

  return (1);
} /* x86DisableBreakpoint() */

/*
x86DumpMemory()
  Dump memory contents of debugged process

Inputs: buf   - buffer to store memory bytes in
        start - address to start dump
        bytes - number of bytes to dump

Return: number of bytes dumped - if this is less than 'bytes',
        an error occurred, and errno should be set appropriately.

Side effects: space is allocated for 'buf', so it must be freed by
              the calling function
*/

long
x86DumpMemory(char **buf, unsigned long start, long bytes)

{
  int wordval;
  unsigned long addr, /* current address we are examining */
                end;  /* last address to examine */
  char *bufptr;
  long ret;           /* return value */

  assert(bytes > 0);

  if (DebugPid == NOPID)
  {
    if (x86StartDebugProcess() == 0)
      return (0); /* something went wrong */
  }

  addr = start;
  end = start + bytes;

  *buf = (char *) MyMalloc(bytes + 1);
  bufptr = *buf;

  ret = 0;

  while (addr < end)
  {
    wordval = PtraceRead(DebugPid, addr, 0);
    if (wordval == (-1))
    {
      /*
       * Most likely input/output error (EIO)
       */
      return (ret);
    }

    /*
     * The ptrace() call will return a 2 byte (word) value on
     * intel systems, and little-endian will put the byte at
     * the address 'addr' last, so anding with 0xff will separate
     * that byte.
     */
    *bufptr++ = (char) (wordval & 0xff);

    ++addr;
    ++ret;
  } /* while (addr <= end) */

  *bufptr = '\0';

  return (ret);
} /* x86DumpMemory() */

/*
x86SetMemory()
  Set the contents of memory location 'address' to 'value'

Return: 0 if unsuccessful (ptrace error)
        number of bytes written if successful
*/

int
x86SetMemory(unsigned long address, unsigned long value)

{
  int wordval;
  int ret;
  unsigned int mask;

  if (DebugPid == NOPID)
  {
    if (x86StartDebugProcess() == 0)
      return (0); /* something went wrong */
  }

  /*
   * An interesting note about ptrace() - although it is documented
   * to return a "word", it actually returns the size of an int,
   * in this case 32 bits, when it should actually return 16 bits,
   * the size of an intel word.
   * This works for PtraceWrite() too - it writes a 32 bit quantity
   * to the address you specify.
   */
  wordval = PtraceRead(DebugPid, address, 0);
  if (wordval == (-1))
    return (0);

  ret = 0;

  /*
   * Determine how many bytes the calling program wants to write
   * to memory - if we simply try to write the 32 bit quantity
   * 'value', and it is a 1 byte number, we will overwrite 3 bytes
   * of memory with zeros
   */
  if ((value & 0xff) == value)
  {
    ret = 1; /* we are setting 1 byte */
    mask = 0xffffff00;
  }
  else if ((value & 0xffff) == value)
  {
    ret = 2; /* we are setting 2 bytes */
    mask = 0xffff0000;
  }
  else if ((value & 0xffffff) == value)
  {
    ret = 3; /* we are setting 3 bytes */
    mask = 0xff000000;
  }
  else
  {
    ret = 4; /* we are setting 4 bytes */
    mask = 0x00000000;
  }

  /*
   * Clear the byte(s) we are about to change
   */
  wordval &= mask;

  /*
   * Set the new byte(s)
   */
  wordval |= value;

  if (PtraceWrite(DebugPid, address, wordval) != 0)
    return (0);

  return (ret);
} /* x86SetMemory() */
