/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: elf.c,v 1.10 2001/04/07 03:04:32 cosine Exp $
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "alddefs.h"
#include "disassemble.h"
#include "elfdefs.h"
#include "load.h"
#include "print.h"

/*
 * libOp includes
 */
#include "disasm.h"

/*
 * libString includes
 */
#include "Strn.h"

/*
 * The _Elf* functions are internal functions which serve as backends
 * to the global user-friendly functions called outside this module.
 */
static Elf32_Shdr *_ElfFindSectionByName(Elf32_Ehdr *hptr, char *name);

static char *ElfClass[] = {
  "Invalid class",             /* ELFCLASSNONE */
  "32 bit",                    /* ELFCLASS32 */
  "64 bit"                     /* ELFCLASS64 */
};

static char *ElfData[] = {
  "Invalid data",              /* ELFDATANONE */
  "LSB",                       /* ELFDATA2LSB */
  "MSB"                        /* ELFDATA2MSB */
};

static char *ElfType[] = {
  "No file type",              /* ET_NONE */
  "Relocatable",               /* ET_REL */
  "Executable",                /* ET_EXEC */
  "Shared Object",             /* ET_DYN */
  "Core file"                  /* ET_CORE */
};

static char *ElfMachine[] = {
  "No machine",                /* EM_NONE */
  "AT&T WE 32100",             /* EM_M32 */
  "SUN SPARC",                 /* EM_SPARC */
  "Intel 80386",               /* EM_386 */
  "Motorola m68k family",      /* EM_68K */
  "Motorola m88k family",      /* EM_88K */
  "Intel 80486",               /* EM_486 */
  "Intel 80860",               /* EM_860 */
  "MIPS R3000 big-endian",     /* EM_MIPS */
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  "Intel 80960",               /* EM_960 */
  "PowerPC"                    /* EM_PPC */
};

static char *ElfVersion[] = {
  "Invalid version",           /* EV_NONE */
  "Version 1 (current)"        /* EV_CURRENT */
};

static char *SectionType[] = {
  "inactive",                  /* SHT_NULL */
  "program specific",          /* SHT_PROGBITS */
  "symbol table",              /* SHT_SYMTAB */
  "string table",              /* SHT_STRTAB */
  "relocation entries",        /* SHT_RELA */
  "symbol hash table",         /* SHT_HASH */
  "dynamic linking data",      /* SHT_DYNAMIC */
  "note",                      /* SHT_NOTE */
  "no bits",                   /* SHT_NOBITS */
  "relocation entries",        /* SHT_REL */
  "reserved",                  /* SHT_SHLIB */
  "dynamic symbol table",      /* SHT_DYNSYM */
  "processor specific",        /* SHT_LOPROC */
  "processor specific",        /* SHT_HIPROC */
  "lower bound index range",   /* SHT_LOUSER */
  "upper bound index range"    /* SHT_HIUSER */
};

static char *SymbolBinding[] = {
  "local",                     /* STB_LOCAL */
  "global",                    /* STB_GLOBAL */
  "weak",                      /* STB_WEAK */
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "processor specific",        /* STB_LOPROC */
  "processor specific",
  "processor specific"         /* STB_HIPROC */
};

static char *SymbolType[] = {
  "unknown",                   /* STT_NOTYPE */
  "data object",               /* STT_OBJECT */
  "function",                  /* STT_FUNC */
  "section",                   /* STT_SECTION */
  "file",                      /* STT_FILE */
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "processor specific",        /* STT_LOPROC */
  "processor specific",
  "processor specific"         /* STT_HIPROC */
};

/*
CheckElf()
 Called from LoadFile() to determine if the newly mapped memory
is an Executable and Linkable Format (ELF) file.

Inputs: filename - name of file
        entry    - set to program's entry point if it is an ELF file

Return: 1 if file is an ELF
        0 if not

Side effects: VirtualFileAddress, VirtualEntryPoint and EntryPoint
              are set to appropriate values if the file is an ELF
              file
*/

int
CheckElf(char *filename, unsigned int *entry)

{
  Elf32_Ehdr *ElfHeader;
  Elf32_Phdr *ProgramHeader;

  assert(filename != 0);
  assert(MapPtr != 0);

  /*
   * Make sure the file is at least as big as the ELF header
   */
  if (MappedSize < sizeof(Elf32_Ehdr))
    return (0);

  ElfHeader = MapPtr;

  if ((ElfHeader->e_ident[EI_MAG0] != ELFMAG0) ||
      (ElfHeader->e_ident[EI_MAG1] != ELFMAG1) ||
      (ElfHeader->e_ident[EI_MAG2] != ELFMAG2) ||
      (ElfHeader->e_ident[EI_MAG3] != ELFMAG3))
  {
    /*
     * Invalid ELF header
     */
    return (0);
  }

  if (ElfHeader->e_ident[EI_CLASS] >= sizeof(ElfClass))
    return (0);

  if (ElfHeader->e_ident[EI_DATA] >= sizeof(ElfData))
    return (0);

  if (ElfHeader->e_type >= sizeof(ElfType))
    return (0);

  if (ElfHeader->e_machine >= sizeof(ElfMachine))
    return (0);

  if (ElfHeader->e_version >= sizeof(ElfVersion))
    return (0);

  assert(ElfMachine[(unsigned char) ElfHeader->e_machine] != 0);

  Print(P_COMMAND,
    "%s: ELF %s (%s), %s, %s, %s",
    filename,
    ElfMachine[(unsigned char) ElfHeader->e_machine],
    ElfClass[(unsigned char) ElfHeader->e_ident[EI_CLASS]],
    ElfData[(unsigned char) ElfHeader->e_ident[EI_DATA]],
    ElfType[(unsigned char) ElfHeader->e_type],
    ElfVersion[(unsigned char) ElfHeader->e_version]);

  /**entry = ElfHeader->e_entry;*/
  VirtualEntryPoint = ElfHeader->e_entry;

  ProgramHeader = (Elf32_Phdr *) ((char *)ElfHeader + ElfHeader->e_phoff);

  if (ProgramHeader->p_vaddr != 0)
  {
    VirtualFileAddress = (unsigned int) (ProgramHeader->p_vaddr - ProgramHeader->p_offset);

    EntryPoint = VirtualEntryPoint - VirtualFileAddress;
  }
  else
    VirtualFileAddress = VirtualEntryPoint = EntryPoint = 0;

  /*
   * When we start disassembling the file, we will want to
   * use 32 bit data and addressing
   */
  DisassembleFlags |= OP_BITS32;

  return (1);
} /* CheckElf() */

/*
PrintElfHeader()
  Print detailed information about the elf header 'hptr'
*/

void
PrintElfHeader(Elf32_Ehdr *hptr)

{
  Elf32_Phdr *phdr;
  unsigned int entryoff;

  assert(hptr != 0);

  /*
   * We can assume we have a valid ELF header
   */
  Print(P_COMMAND, "%s:", ObjectFileName);

  Print(P_COMMAND,
    "Executable and Linkable Format (ELF), %s, %s, %s",
    ElfClass[(unsigned char) hptr->e_ident[EI_CLASS]],
    ElfData[(unsigned char) hptr->e_ident[EI_DATA]],
    ElfVersion[(unsigned char) hptr->e_version]);

  Print(P_COMMAND,
    "Machine Architecture:               %s",
    ElfMachine[(unsigned char) hptr->e_machine]);

  Print(P_COMMAND,
    "Object File Type:                   %s",
    ElfType[(unsigned char) hptr->e_type]);

  phdr = (Elf32_Phdr *) ((char *)hptr + hptr->e_phoff);
  entryoff = hptr->e_entry - phdr->p_vaddr + phdr->p_offset;

  Print(P_COMMAND,
    "Virtual Entry Point:                0x%08X (File Offset: 0x%08X)",
    hptr->e_entry,
    entryoff);

  Print(P_COMMAND,
    "Program table header offset:        0x%08X",
    hptr->e_phoff);

  Print(P_COMMAND,
    "Section table header offset:        0x%08X",
    hptr->e_shoff);

  Print(P_COMMAND,
    "Processor specific flags:           0x%08X",
    hptr->e_flags);

  Print(P_COMMAND,
    "ELF header size (bytes):            0x%04X",
    hptr->e_ehsize);

  Print(P_COMMAND,
    "Program header table entry size:    0x%04X",
    hptr->e_phentsize);

  Print(P_COMMAND,
    "Program header table entry count:   0x%04X",
    hptr->e_phnum);

  Print(P_COMMAND,
    "Section header table entry size:    0x%04X",
    hptr->e_shentsize);

  Print(P_COMMAND,
    "Section header table entry count:   0x%04X",
    hptr->e_shnum);

  Print(P_COMMAND,
    "Section header string table index:  0x%04X",
    hptr->e_shstrndx);
} /* PrintElfHeader() */

/*
PrintElfSectionInfo()
  Print the addresses and names of the sections of the elf
file

Inputs: hptr  - pointer to elf header
        sname - optional section name (if not given, all sections
                will be displayed)
*/

void
PrintElfSectionInfo(Elf32_Ehdr *hptr, char *sname)

{
  char *strtable;
  Elf32_Shdr *sectable;
  Elf32_Shdr *sptr;
  Elf32_Word size;
  Elf32_Half ii; /* looping */

  assert (hptr != 0);

  if (hptr->e_shnum == 0)
  {
    Print(P_COMMAND, "No sections found");
    return;
  }

  /*
   * Get the location of the section header table
   */
  sectable = (Elf32_Shdr *) ((char *) hptr + hptr->e_shoff);

  /*
   * The location of the string table is inside one of the
   * structures in the section header table, given by the index
   * e_shstrndx (ie: sectable + hptr->e_shstrndx*sizeof(Elf32_Shdr))
   */
  sptr = sectable + hptr->e_shstrndx;
  assert(sptr->sh_type == SHT_STRTAB);

  strtable = (char *) hptr + sptr->sh_offset;

  if (sname)
  {
    char fbuf[MAXLINE]; /* flags buffer */
    int flen;

    sptr = _ElfFindSectionByName(hptr, sname);
    if (!sptr)
    {
      Print(P_COMMAND,
            "No section found matching: %s",
            sname);
      return;
    }

    Print(P_COMMAND,
          "Section name:      %s",
          strtable + sptr->sh_name);
    Print(P_COMMAND,
          "Section type:      %s",
          SectionType[sptr->sh_type]);

    fbuf[0] = '\0';
    if (sptr->sh_flags & SHF_WRITE)
      strcat(fbuf, "Writable, ");
    if (sptr->sh_flags & SHF_ALLOC)
      strcat(fbuf, "Occupies memory during execution, ");
    if (sptr->sh_flags & SHF_EXECINSTR)
      strcat(fbuf, "Contains executable instructions, ");
    if (sptr->sh_flags & SHF_MASKPROC)
      strcat(fbuf, "Reserved for processor-specific semantics, ");

    flen = strlen(fbuf);
    if (flen && (fbuf[flen - 2] == ','))
      fbuf[flen - 2] = '\0';

    if (*fbuf)
    {
      Print(P_COMMAND,
            "Section flags:     %s",
            fbuf);
    }

    Print(P_COMMAND,
          "Virtual address:   0x%08X",
          sptr->sh_addr);
    Print(P_COMMAND,
          "Size (bytes):      0x%08X (%d)",
          sptr->sh_size,
          sptr->sh_size);
    Print(P_COMMAND,
          "File offset:       0x%08X (%d)",
          sptr->sh_offset,
          sptr->sh_offset);

    if ((sptr->sh_addralign != 0) && (sptr->sh_addralign != 1))
    {
      Print(P_COMMAND,
            "Address alignment: %d bytes",
            sptr->sh_addralign);
    }

    if (sptr->sh_entsize != 0)
    {
      Print(P_COMMAND,
            "Entry size:        %d bytes",
            sptr->sh_entsize);
    }

    return;
  } /* if (sname) */

  /*
   * They did not give a specific section name - output all
   * sections
   */

  PrintBurstStart();

  Print(P_COMMAND,
    "%-20s %-8s      %-8s      %-8s",
    "Section name",
    "Start",
    "End",
    "Length (bytes)");

  for (ii = 0; ii < hptr->e_shnum; ++ii)
  {
    sptr = sectable + ii;
    if (sptr->sh_type == SHT_NULL)
      continue; /* unused section */

    /*
     * The ELF specification says that sh_size can be non-zero if
     * the section is marked as SHT_NOBITS, which means that the
     * section has zero size, so I will special-case this
     */
    if (sptr->sh_type == SHT_NOBITS)
      size = 0;
    else
      size = sptr->sh_size;

    Print(P_COMMAND,
      "%-20s 0x%08X -> 0x%08X    0x%08X",
      strtable + sptr->sh_name,
      VirtualFileAddress + sptr->sh_offset,
      VirtualFileAddress + sptr->sh_offset + size,
      size);
  }

  PrintBurstEnd();
} /* PrintElfSectionInfo() */

/*
FindElfSectionByAddr()
  Find elf section containing given address

Inputs: address - file address
        start   - modified to contain start of segment
        end     - modified to contain end of segment

Return: 1 if section found
        0 if not

Note: address must be given in virtual form
*/

int
FindElfSectionByAddr(Elf32_Ehdr *hptr, unsigned int address,
                     unsigned int *start, unsigned int *end)

{
  Elf32_Shdr *sectable;
  Elf32_Shdr *sptr;
  Elf32_Half ii; /* looping */
  Elf32_Off begin,
            finish;

  assert(hptr && start && end);

  if (hptr->e_shnum == 0)
    return (0); /* no sections */

  /*
   * Get the location of the section header table
   */
  sectable = (Elf32_Shdr *) ((char *) hptr + hptr->e_shoff);

  for (ii = 0; ii < hptr->e_shnum; ++ii)
  {
    sptr = sectable + ii;

    begin = VirtualFileAddress + sptr->sh_offset;
    finish = begin + sptr->sh_size;

    if ((address >= begin) && (address <= finish))
    {
      /*
       * We have found the section
       */
      *start = begin;
      *end = finish;
      return (1);
    }
  }

  return (0);
} /* FindElfSectionByAddr() */

/*
FindElfSectionByName()
  Find elf section by name

Inputs: name  - section name
        start - modified to contain start of segment
        end   - modified to contain end of segment

Return: 1 if section found
        0 if not
*/

int
FindElfSectionByName(Elf32_Ehdr *hptr, char *name,
                     unsigned int *start, unsigned int *end)

{
  Elf32_Shdr *sptr;

  assert(hptr && name && start && end);

  sptr = _ElfFindSectionByName(hptr, name);

  if (!sptr)
    return (0); /* section not found */

  /*
   * We found the section
   */
  *start = VirtualFileAddress + sptr->sh_offset;
  *end = *start + sptr->sh_size;

  return (1);
} /* FindElfSectionByName() */

/*
_ElfFindSectionByName()
  Backend to FindElfSectionByName() - locate elf section with the
give name, if any and return a pointer to it

Inputs: hptr - pointer to elf header
        name - section name

Return: pointer to section if found, 0 if not
*/

static Elf32_Shdr *
_ElfFindSectionByName(Elf32_Ehdr *hptr, char *name)

{
  Elf32_Shdr *sectable;
  Elf32_Shdr *sptr;
  Elf32_Half ii; /* looping */
  char *strtable;
  char *sname;

  assert(hptr && name);

  if (hptr->e_shnum == 0)
    return (0); /* no sections */

  /*
   * Get the location of the section header table
   */
  sectable = (Elf32_Shdr *) ((char *) hptr + hptr->e_shoff);

  /*
   * Get the location of the string header table
   */
  strtable = (char *) ((char *) hptr + (sectable + hptr->e_shstrndx)->sh_offset);

  for (ii = 0; ii < hptr->e_shnum; ++ii)
  {
    sptr = sectable + ii;

    sname = strtable + sptr->sh_name;

    if (!Strcasecmp(name, sname))
    {
      /*
       * match found
       */
      return (sptr);
    }
  }

  return (0);
} /* _ElfFindSectionByName() */

/*
PrintElfSymbols()
  Print object file's symbols if any
*/

void
PrintElfSymbols(Elf32_Ehdr *hptr)

{
  Elf32_Shdr *sectable; /* pointer to section header table */
  Elf32_Shdr *sptr;     /* temporary section header table entry */
  Elf32_Sym *symtabptr, /* pointer to symbol table */
            *symptr;    /* temp pointer */
  char *strtabptr;      /* pointer to symbol string table */
  Elf32_Half ii;        /* looping */
  unsigned long symcnt; /* number of entries in the symbol table */
  char addrstr[64];

  assert (hptr != 0);

  if (hptr->e_shnum == 0)
  {
    /*
     * No sections => no symbols
     */
    Print(P_COMMAND, "No symbols found");
    return;
  }

  /*
   * Get the location of the section header table
   */
  sectable = (Elf32_Shdr *) ((char *) hptr + hptr->e_shoff);

  strtabptr = 0;
  symtabptr = 0;
  symcnt = 0;

  for (ii = 0; ii < hptr->e_shnum; ++ii)
  {
    sptr = sectable + ii;

#if 0
    if (sptr->sh_type == SHT_SYMTAB)
    {
      Print(P_COMMAND,
            "Found a symtab section: 0x%08X flags=%d size=%d link=%d info=%d addralign=%d entsize=%d offset=%d",
            VirtualFileAddress + sptr->sh_offset,
            sptr->sh_flags,
            sptr->sh_size,
            sptr->sh_link,
            sptr->sh_info,
            sptr->sh_addralign,
            sptr->sh_entsize,
            sptr->sh_offset);
    }

    if (sptr->sh_type == SHT_STRTAB)
    {
      Print(P_COMMAND,
            "Found a STRtab section: 0x%08X flags=%d size=%d link=%d info=%d addralign=%d entsize=%d",
            VirtualFileAddress + sptr->sh_offset,
            sptr->sh_flags,
            sptr->sh_size,
            sptr->sh_link,
            sptr->sh_info,
            sptr->sh_addralign,
            sptr->sh_entsize);
    }
#endif

    if (sptr->sh_type == SHT_SYMTAB)
    {
      /*
       * We found a symbol section (.symtab) - there should only
       * be one section with the property of SHT_SYMTAB, so if
       * symtabptr is non-null we have a problem.
       */
      assert(!symtabptr);
      symtabptr = (Elf32_Sym *) ((char *) hptr + sptr->sh_offset);
      symcnt = sptr->sh_size / sptr->sh_entsize;

      /*
       * The string table associated with the symbol table is
       * the section header table index specified by the sh_link field
       */
      strtabptr = (char *) ((char *) hptr +
                            (sectable + sptr->sh_link)->sh_offset);
    }
  } /* for (ii = 0; ii < hptr->e_shnum; ++ii) */

  if (!symtabptr)
  {
    Print(P_COMMAND, "No symbols found");
    return;
  }

  PrintBurstStart();

  Print(P_COMMAND,
        "%-10s %-10s %-20s %-30s %s",
        "Address",
        "Binding",
        "Type",
        "Name",
        "Size");

  /*
   * Skip the first symbol table entry since it is defined to be
   * a dummy entry
   */
  for (ii = 1; ii < symcnt; ++ii)
  {
    symptr = symtabptr + ii;

    if (symptr->st_value != 0)
      sprintf(addrstr, "0x%08X", symptr->st_value);
    else
      *addrstr = 0;

    Print(P_COMMAND,
          "%-10s %-10s %-20s %-30s %d",
          addrstr,
          SymbolBinding[ELF32_ST_BIND(symptr->st_info)],
          SymbolType[ELF32_ST_TYPE(symptr->st_info)],
          strtabptr + symptr->st_name,
          symptr->st_size);
  }

  PrintBurstEnd();
} /* PrintElfSymbols() */
