/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: load.c,v 1.3 2000/09/22 20:14:59 cosine Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#include "elfdefs.h"
#include "load.h"
#include "misc.h"
#include "print.h"

/*
 * libDebug includes
 */
#include "debug.h"

/*
 * libString includes
 */
#include "alloc.h"
#include "Strn.h"

static int IdentifyFileType(char *filename, unsigned int *entry);

/*
 * Global: pointer to mapped memory
 */
void                 *MapPtr = 0;

/*
 * Global: size of mapped memory
 */
size_t               MappedSize = 0;

/*
 * Global: file we are debugging
 */
int                  ObjectFileDescriptor = (-1);

/*
 * Global: name of file we are debugging
 */
char                 *ObjectFileName = 0;

/*
 * Global: object format of file we are debugging (OT_xxx)
 */
int                  ObjectFileFormat = OT_UNKNOWN;

/*
 * Global: current file offset of file we are debugging
 */
unsigned int         ObjectFileOffset = 0;

/*
 * Global: entry point of file we are debugging
 */
unsigned int         EntryPoint = 0;

/*
 * Global: virtual address of the beginning of the file we are
 *         debugging
 */
unsigned int         VirtualFileAddress = 0;

/*
 * Global: virtual address of the program's entry point
 */
unsigned int         VirtualEntryPoint = 0;

/*
 * Global: virtual address of the file's current offset
 */
unsigned int         VirtualObjectFileOffset = 0;

/*
LoadFile()
 Load file 'filename' by using mmap() to create a map of the file
in memory. This is the most efficient method since it also pages
for us.

Return: -1 if unsuccessful
        file descriptor of 'filename' is successful

Side effects:
 - MapPtr is assigned to the beginning of the mapped memory
 - MappedSize is set to the size of the mapped memory
*/

int
LoadFile(char *filename)

{
  struct stat statbuf;
  unsigned int entry; /* entry point */
  int fd;

  assert(filename != 0);

  /*
   * Unload old file if there is one
   */
  if (MapPtr != 0)
    UnloadFile(filename);

  if ((fd = open(filename, O_RDONLY)) == (-1))
  {
    Print(P_COMMAND,
      "Unable to open %s: %s",
      filename,
      strerror(errno));
    return (-1);
  }

  if (stat(filename, &statbuf) == (-1))
  {
    Print(P_COMMAND,
      "stat() failed on %s: %s",
      filename,
      strerror(errno));
    close(fd);
    return (-1);
  }

  MappedSize = statbuf.st_size;

  MapPtr = mmap(0, MappedSize, PROT_READ, MAP_PRIVATE, fd, 0);
  if (MapPtr == MAP_FAILED)
  {
    Print(P_COMMAND,
      "Unable to map memory for %s: %s",
      filename,
      strerror(errno));
    close(fd);
    return (-1);
  }

  if ((ObjectFileFormat = IdentifyFileType(filename, &entry)) == OT_UNKNOWN)
  {
    Print(P_COMMAND,
      "%s: Unrecognized object format, placing offset at position 0x%08X",
      filename,
      0);
    ObjectFileOffset = 0;
    VirtualObjectFileOffset = 0;
    EntryPoint = 0;
    VirtualEntryPoint = 0;
  }
  else
  {
    /*
     * We now want to set the file offset to the program's entry
     * point. EntryPoint and VirtualEntryPoint should already be
     * set.
     */
    ObjectFileOffset = EntryPoint;
    VirtualObjectFileOffset = VirtualEntryPoint;
  }

  ObjectFileName = Strdup(filename);

  /*
   * Get the file ready for debugging
   */
  InitializeDebugProcess(ObjectFileName, ModeCurses);

  return (fd);
} /* LoadFile() */

/*
UnloadFile()
 Unmap the previously mapped memory for the file we were debugging.
*/

void
UnloadFile(char *filename)

{
  assert(MapPtr != 0);
  assert(filename != 0);

  if (munmap(MapPtr, MappedSize) == (-1))
  {
    Print(P_COMMAND,
      "Error unloading file %s: %s",
      filename,
      strerror(errno));
    return;
  }

  if (ObjectFileName)
    MyFree(ObjectFileName);

  MapPtr = 0;
} /* UnloadFile() */

/*
IdentifyFileType()
 Attempt to identify the object format of 'filename'

Return: OT_xxx if match found
        OT_UNKNOWN if no match found
*/

static int
IdentifyFileType(char *filename, unsigned int *entry)

{
  assert(filename != 0);

  if (CheckElf(filename, entry))
    return (OT_ELF);

  return (OT_UNKNOWN);
} /* IdentifyFileType() */
