/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: main.c,v 1.10 2001/04/07 18:53:18 cosine Exp $
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "alddefs.h"
#include "command.h"
#include "defs.h"
#include "load.h"
#include "misc.h"
#include "signals.h"
#include "terminal.h"
#include "version.h"
#include "window.h"

/*
 * libString includes
 */
#include "Strn.h"

static int ParseCommandLine(int ac, char *av[]);

extern char libDebugVersion[];
extern char libOpVersion[];

/*
 * Global: Use console mode
 */
int                        ModeConsole = 0;

/*
 * Global: Use ncurses mode
 */
int                        ModeCurses = 0;

/*
 * Global: address of the next instruction we will execute
 */
unsigned long              NextInstruction = 0;

/*
 * Local: filename specified on command line if any
 */
char                       *Filename = 0;

/*
ParseCommandLine()
 Parse command line arguments

 Inputs: ac - argument count
         av - argument array

 Return: 1 upon success
         0 upon failure
*/

static int
ParseCommandLine(int ac, char *av[])

{
  while (--ac)
  {
    if (*av[ac] == '-')
    {
      switch (*(av[ac] + 1))
      {
        /*
         * Give help output
         */
        case 'h':
        {
          fprintf(stdout,
            "Usage: %s [options] [filename]\n\
\n\
  [filename]  : Path to executable file to debug\n\
\n\
Options:\n\
\n\
  -c          : Use curses display (if available)\n\
  -h          : Output this help screen\n\
  -v          : Output version information\n",
            av[0]);

          MyExit(0);

          break; /* not reached */
        } /* case 'h' */

        /*
         * Give version information
         */
        case 'v':
        {
          fprintf(stdout,
            "ald version:       %s\n",
            aVersion);
          fprintf(stdout,
            "libDebug version:  %s\n",
            libDebugVersion);
          fprintf(stdout,
            "libOp version:     %s\n",
            libOpVersion);
          fprintf(stdout,
            "libString version: %s\n",
            libStringVersion);

          MyExit (0);

          break; /* not reached */
        } /* case 'v' */

        /*
         * Use curses
         */
        case 'c':
        {
          ModeCurses = 1;
          break;
        } /* case 'c' */
      }
    }
    else
    {
      /*
       * No "-" switch, must be the filename
       */
      Filename = av[ac];
    }
  }

  if (!ModeCurses)
    ModeConsole = 1;

  return (1);
} /* ParseCommandLine() */

int
main(int argc, char *argv[])

{
  fprintf(stdout, "Assembly Language Debugger %s\n",
    aVersion);
  fprintf(stdout, "Copyright (C) 2000 Patrick Alken\n\n");

  if (!ParseCommandLine(argc, argv))
    MyExit (0);

  SetupSignals();

  InitTerm();

  CmdPrompt = Strdup(ALD_PROMPT);

  if (ModeConsole)
  {
    /*
     * Save our terminal attributes
     */
    SaveParentTerminalState();

    /*
     * This is just to initialize ChildAttributes
     */
    SaveChildTerminalState();
  }
  else if (ModeCurses)
  {
  #ifdef USE_CURSES

    fprintf(stderr, "ncurses support is currently disabled until a future release\n");
    exit(1);

    if (!InitWindows())
      MyExit(1);

  #else

    fprintf(stderr,
      "This binary was compiled without ncurses support\n");
    MyExit (0);

  #endif /* !USE_CURSES */
  }

  if (Filename)
    ObjectFileDescriptor = LoadFile(Filename);

  /*
   * Start processing commands
   */
  GetCommandInput();

  return (0);
} /* main() */
