/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: memory.c,v 1.4 2001/04/07 01:00:01 cosine Exp $
 */

#include <stdlib.h>
#include <assert.h>

#include "alddefs.h"
#include "defs.h"
#include "memory.h"
#include "print.h"

struct MemFormat OutputFormats[] = {
  { 'x', "hex", "%02X ", "   ", 16 },         /* MF_HEX */
  { 'o', "octal", "%03o ", "    ", 8 },       /* MF_OCTAL */
  { 'd', "decimal", "%03d ", "    ", 8 },     /* MF_DECIMAL */

  { 0, 0, 0, 0 }
};

static struct MemFormat *GetOutputFormat(unsigned char output);

/*
GetOutputFormat()
  Look up a suitable output format for the given output arg.

Inputs: output - character representing a base output

Return: pointer to index of OutputFormats[] corresponding to 'output'.
        If 'output' does not match any known formats, default to
        hexadecimal.
*/

static struct MemFormat *
GetOutputFormat(unsigned char output)

{
  if (output == 'o')
    return (&OutputFormats[MF_OCTAL]);
  else if (output == 'd')
    return (&OutputFormats[MF_DECIMAL]);

  return (&OutputFormats[MF_HEX]);
} /* GetOutputFormat() */

/*
GetElementSize()
  This routine calculates the numerical size of each element
being examined in memory as a result of the "examine -size X"
command.

Inputs: str - string containing size

Return: positive value corresponding to the number of bytes in 'str'.
        If str contains a positive, non-zero value, that is converted
        to an integer and returned. If it contains a special letter,
        that is converted to an integer according to the following
        rules.
        If str contains a negative number or an invalid letter,
        return 0.

letter         value      meaning
b              1          byte
h              ? / 2      halfword
w              ?          word
d              ? * 2      doubleword
g              8          giant (8 bytes)

? - This value depends on the size of the machine word for the
    architecture we are running on
*/

long
GetElementSize(char *str)

{
  long ret;
  char *endptr;
  unsigned char c;

  assert(str != 0);

  ret = strtol(str, &endptr, 0);
  if ((endptr != str) && (*endptr == '\0'))
  {
    /*
     * It is a valid number
     */
    if (ret > 0)
      return (ret);
    else
      return (0); /* negative number - no good */
  }

  c = (unsigned char) *str;
  switch (c)
  {
    /*
     * One byte
     */
    case 'b':
    case 'B':
    {
      return (1);
      /*NOTREACHED*/
      break;
    }

    /*
     * One word
     */
    case 'w':
    case 'W':
    {
      return (WORDSIZE);
      /*NOTREACHED*/
      break;
    }

    /*
     * Half a word
     */
    case 'h':
    case 'H':
    {
      return (HALFWORD);
      /*NOTREACHED*/
      break;
    }

    /*
     * Double word
     */
    case 'd':
    case 'D':
    {
      return (DWORD);
      /*NOTREACHED*/
      break;
    }

    /*
     * Giant (8 bytes)
     */
    case 'g':
    case 'G':
    {
      return (8);
      /*NOTREACHED*/
      break;
    }

    /*
     * Unknown size
     */
    default:
    {
      return (0);
      /*NOTREACHED*/
      break;
    }
  } /* switch (c) */

  /*NOTREACHED*/
  return (0);
} /* GetElementSize() */

/*
GetMemoryValue()
  Determine the numerical value of the next element in memory

Inputs: buf  - buffer containing next element
        size - size (in bytes) of next element

Return: value of next element of memory, -1 if unsuccessful
*/

long
GetMemoryValue(unsigned char *buf, long size)

{
  long ret = (-1);
  int length = 0;

  /*
   * Thank god for little endian :-)
   */

  if (size == 1)
  {
    ret = (unsigned char) buf[length++];
  }
  else if (size == 2)
  {
    ret = (unsigned char) buf[length++];
    ret += (unsigned char) buf[length++] * 256;
  }
  else if (size == 4)
  {
    ret = (unsigned char) buf[length++];
    ret += (unsigned char) buf[length++] * 256;
    ret += (unsigned char) buf[length++] * 65536;
    ret += (unsigned char) buf[length++] * 16777216;
  }

  return (ret);
} /* GetMemoryValue() */

/*
OutputMemory()
  Output a memory dump in a readable format

Inputs: buf    - buffer containing bytes of memory
        start  - address of first byte in memory
        bytes  - number of bytes in 'buf'
        size   - size of each element in 'buf'
        output - output format (hex, dec, etc)
*/

void
OutputMemory(unsigned char *buf, unsigned long start,
             long bytes, long size, unsigned char output)

{
  unsigned char *bufptr;
  long element;        /* current element we are printing */
  unsigned char *end;
  struct MemFormat *fptr;
  unsigned long addr;  /* current address we are printing */
  char abuf[MAXLINE];  /* ascii buffer */
  char *aptr;
  int bcnt;            /* number of bytes we have printed so far */
  int extranl;
  int ii;

  fptr = GetOutputFormat(output);
  assert(fptr != 0);

  end = buf + bytes;
  bufptr = buf;

  /*PrintBurstStart();*/

  Print(P_MEMORY,
    "Dumping %ld bytes of memory starting at 0x%08lX in %s",
    bytes,
    start,
    fptr->desc);

  bcnt = 0;
  addr = start;
  aptr = abuf;
  extranl = 1;

  RawPrint(P_MEMORY, "%08lX:  ", start);

  while (bufptr < end)
  {
    element = GetMemoryValue(bufptr, size);
    if (element < 0)
    {
      /*
       * Should not happen
       */
      Print(P_ERROR, "GetMemoryValue() failed for size %ld", size);
      /*PrintBurstEnd();*/
      return;
    }

    RawPrint(P_MEMORY, fptr->fmt, *bufptr);

    addr += size;

    /*
     * Keep our ascii buffer updated, as well as advance bufptr
     * by 'size' bytes
     */
    for (ii = 0; ii < size; ++ii)
      *aptr++ = *bufptr++;

    if ((++bcnt == fptr->bpl) || (bufptr >= end))
    {
      if (bufptr >= end)
      {
        /*
         * We have partially completed the last line - fill the
         * rest up with spaces so we can line our ascii printout
         * up correctly.
         */
        for (ii = bcnt; ii < fptr->bpl; ++ii)
          RawPrint(P_MEMORY, fptr->blank);
      }

      RawPrint(P_MEMORY, "   ");

      /*
       * Output the ascii equivalent of the bytes we just
       * printed
       */
      aptr = abuf;
      for (ii = 0; ii < bcnt; ++ii)
      {
        if ((*(aptr + ii) >= ' ') && (*(aptr + ii) < 127))
          RawPrint(P_MEMORY, "%c", (unsigned char) *(aptr + ii));
        else
          RawPrint(P_MEMORY, "%c", '.');
      }

      RawPrint(P_MEMORY, "\n");
      if (bufptr < end)
        RawPrint(P_MEMORY, "%08lX:  ", addr);
      else
        extranl = 0;

      bcnt = 0;
    }
  }

  if (extranl)
    RawPrint(P_MEMORY, "\n");

  /*PrintBurstEnd();*/
} /* OutputMemory() */
