/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: set.c,v 1.5 2001/04/07 18:53:18 cosine Exp $
 */

#include <stdlib.h>

#include "alddefs.h"
#include "command.h"
#include "load.h"
#include "misc.h"
#include "msg.h"
#include "print.h"
#include "set.h"

/*
 * libDebug includes
 */
#include "args.h"

/*
 * libString includes
 */
#include "alloc.h"
#include "Strn.h"

static void SetArgs(int ac, char **av, unsigned int pwin);
static void SetEntryPoint(int ac, char **av, unsigned int pwin);
static void SetFileOffset(int ac, char **av, unsigned int pwin);
static void SetPausePrint(int ac, char **av, unsigned int pwin);
static void SetPrompt(int ac, char **av, unsigned int pwin);

struct Command setcmds[] = {
  { "args", SetArgs, 0 },
  { "entry-point", SetEntryPoint, 0 },
  { "file-offset", SetFileOffset, 0 },
  { "offset", SetFileOffset, C_ALIAS },
  { "pause-print", SetPausePrint, 0 },
  { "prompt", SetPrompt, 0 },
  { 0, 0, 0 }
};

/*
 * List of setting variables and their default values
 */

/*
DisplaySettings()
  Output a list of all settings and their values
*/

void
DisplaySettings()

{
  struct Command *cptr;

  for (cptr = setcmds; cptr->cmd != 0; ++cptr)
  {
    if (cptr->flags & C_ALIAS)
      continue; /* don't display aliases */

    RawPrint(P_COMMAND, "%-30s", cptr->cmd);
    (*cptr->funcptr)(0, 0, P_COMMAND);
    RawPrint(P_COMMAND, "\n");
  }
} /* DisplaySettings() */

/*
SetArgs()
  Set the runtime arguments for the debugged process
*/

static void
SetArgs(int ac, char **av, unsigned int pwin)

{
  char str[MAXLINE];
  char *tmp;
  int ii;
  int len;

  if (pwin != 0)
  {
    tmp = GetProcessArguments();
    RawPrint(pwin, "%s", tmp ? tmp : "none");
    return;
  }

  if (ac < 3)
  {
    /*
     * No arguments means unset previous args
     */
    SetProcessArguments(0);
    return;
  }

  len = sizeof(str);

  tmp = str;
  for (ii = 2; ii < ac; ++ii)
  {
    tmp += Snprintf(tmp, len, "%s ", av[ii]);
    len = sizeof(str) - (int) (tmp - str);
  }

  SetProcessArguments(str);
} /* SetArgs() */

/*
SetEntryPoint()
  Sets the entry point for the program being debugged
*/

static void
SetEntryPoint(int ac, char **av, unsigned int pwin)

{
  long address;

  if (pwin != 0)
  {
    RawPrint(pwin, "0x%08X (0x%08X)",
      EntryPoint,
      VirtualEntryPoint);
    return;
  }

  if (ac < 3)
  {
    Print(P_COMMAND, "Syntax: set entry-point <address>");
    return;
  }

  address = strtol(av[2], 0, 16);

  EntryPoint = (unsigned int) address;
  VirtualEntryPoint = VirtualFileAddress + (unsigned int) address;
} /* SetEntryPoint() */

/*
SetFileOffset()
  Sets the offset from the beginning of the file being debugged.
Commands such as "disassemble" will use this value as their
starting point
*/

static void
SetFileOffset(int ac, char **av, unsigned int pwin)

{
  long address;

  if (pwin != 0)
  {
    RawPrint(pwin, "0x%08X (0x%08X)",
      ObjectFileOffset,
      VirtualObjectFileOffset);
    return;
  }

  if (ac < 3)
  {
    Print(P_COMMAND, "Syntax: set file-offset <address>");
    return;
  }

  address = strtol(av[2], 0, 16);

  ObjectFileOffset = (unsigned int) address;
  VirtualObjectFileOffset = VirtualFileAddress + (unsigned int) address;
} /* SetFileOffset() */

/*
SetPausePrint()
  Enables/Disables pausing of print bursts for commands which output
a lot of information
*/

static void
SetPausePrint(int ac, char **av, unsigned int pwin)

{
  if (pwin != 0)
  {
    RawPrint(pwin, "%s",
      PausePrint ? "on" : "off");
    return;
  }

  if (ac < 3)
  {
    Print(P_COMMAND, "Syntax: set pause-print <on | off>");
    return;
  }

  PausePrint = StrToBool(av[2]);
} /* SetPausePrint() */

/*
SetPrompt()
  Set the command prompt string
*/

static void
SetPrompt(int ac, char **av, unsigned int pwin)

{
  if (pwin != 0)
  {
    RawPrint(pwin, "\"%s\"",
      CmdPrompt);
    return;
  }

  if (ac < 3)
  {
    Print(P_COMMAND, "Syntax: set prompt <new prompt>");
    return;
  }

  MyFree(CmdPrompt);
  CmdPrompt = Strdup(av[2]);
} /* SetPrompt() */
