/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: terminal.c,v 1.3 2001/04/07 01:00:01 cosine Exp $
 */

#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <curses.h>
#include <term.h>

#include "defs.h"
#include "misc.h"
#include "print.h"

#if defined(HAVE_TERMIOS_H)
#include <termios.h>
#elif defined(HAVE_TERMIO_H)
#include <termio.h>
#endif

/*
 * Global: number of lines per page
 */
int                  LinesPerPage = 0;

/*
 * Global: number of columns per line
 */
int                  ColsPerLine = 0;

#if defined(HAVE_TERMIOS_H)

static int SaveTerminalState(struct termios *state);
static int RestoreTerminalState(struct termios *state);

/*
 * Global: terminal attributes for our debugger process
 */
struct termios               ParentAttributes;

/*
 * Global: terminal attributes for the process we are debugging
 */
struct termios               ChildAttributes;

#elif defined(HAVE_TERMIO_H)

static int SaveTerminalState(struct termio *state);
static int RestoreTerminalState(struct termio *state);

struct termio                ParentAttributes;
struct termio                ChildAttributes;

#endif

/*
InitTerm()
  Sets up some variables for our terminal
*/

void
InitTerm()

{
#ifdef USE_CURSES
  char *termtype;
#endif

  /*
   * Attempt to determine how many lines/cols are in our terminal.
   * Print() makes use of this information for pause-print etc.
   */

  /*
   * Set up defaults
   */
  LinesPerPage = 25;
  ColsPerLine = 80;

#ifdef USE_CURSES

  termtype = getenv("TERM");
  if (termtype)
  {
    /*
     * 2048 should be large enough for all known terminals
     */
    char termbuf[2048];
    int status;

    status = tgetent(termbuf, termtype);
    if (status > 0)
    {
      int lval,
          cval;

      lval = tgetnum("li");
      if (lval >= 0)
        LinesPerPage = lval;

      cval = tgetnum("co");
      if (cval >= 0)
        ColsPerLine = cval;
    }
  }

#endif /* USE_CURSES */
} /* InitTerm() */

/*
SaveTerminalState()
  Save the current state of the terminal into the given variable
*/

#if defined(HAVE_TERMIOS_H)

static int
SaveTerminalState(struct termios *state)

{
  if (tcgetattr(STDIN_FILENO, state) == (-1))
    return (0);

  return (1);
} /* SaveTerminalState() */

static int
RestoreTerminalState(struct termios *state)

{
  if (tcsetattr(STDIN_FILENO, TCSANOW, state) == (-1))
    return (0);

  return (1);
} /* RestoreTerminalState() */

#elif defined(HAVE_TERMIO_H)

static int
SaveTerminalState(struct termio *state)

{
  if (ioctl(STDIN_FILENO, TCGETA, state) == (-1))
    return (0);

  return (1);
} /* SaveTerminalState() */

static int
RestoreTerminalState(struct termio *state)

{
  if (ioctl(STDIN_FILENO, TCSETA, state) == (-1))
    return (0);

  return (1);
} /* RestoreTerminalState() */

#endif

/*
SaveParentTerminalState()
  Save the attributes of the terminal into the variable
ParentAttributes

Return: 1 if successful
        0 if not
*/

void
SaveParentTerminalState()

{
  if (!SaveTerminalState(&ParentAttributes))
  {
    Print(P_COMMAND,
      "Error saving parent's terminal state: %s",
      strerror(errno));
    MyExit(1);
  }
} /* SaveParentTerminalState() */

/*
RestoreParentTerminalState()
  Restore the original state of the terminal, using values from
the variable ParentAttributes

Return: 1 if successful
        0 if not
*/

void
RestoreParentTerminalState()

{
  if (!RestoreTerminalState(&ParentAttributes))
  {
    Print(P_COMMAND,
      "Error restoring parent's terminal state: %s",
      strerror(errno));
    MyExit(1);
  }
} /* RestoreParentTerminalState() */

/*
SaveChildTerminalState()
  Save the attributes of the terminal into the variable
ChildAttributes

Return: 1 if successful
        0 if not
*/

void
SaveChildTerminalState()

{
  if (!SaveTerminalState(&ChildAttributes))
  {
    Print(P_COMMAND,
      "Error saving child's terminal state: %s",
      strerror(errno));
    MyExit(1);
  }
} /* SaveChildTerminalState() */

/*
RestoreChildTerminalState()
  Restore the original state of the terminal, using values from
the variable ChildAttributes

Return: 1 if successful
        0 if not
*/

void
RestoreChildTerminalState()

{
  if (!RestoreTerminalState(&ChildAttributes))
  {
    Print(P_COMMAND,
      "Error restoring child's terminal state: %s",
      strerror(errno));
    MyExit(1);
  }
} /* RestoreChildTerminalState() */
